from .classdefine import Addr
from .special import ZXCITIES
from math import radians, cos, sin, asin, sqrt


def lookup(name: str, level: str | None = None) -> str | None:
    """根据地名简称查找全称

    Args:
        name (str): 待查找地名
        level (str | None): 查询区域的行政等级，包括province、city、county三级。默认值为None，当为None时不区分查找范围，因此很可能出现重名错误

    Returns:
        str | None: 地名全称
    """
    if name in ZXCITIES.keys():
        new_name = ZXCITIES[name]
        obj_area = Addr(new_name, level="province").addr
    else:
        obj_area = Addr(name, level=level).addr
    return obj_area[0, "name"]  # type: ignore


def belongs_to(name: str, level: str | None = None) -> str | None:
    """根据地名查找其上级行政区名称

    Args:
        name (str): 待查找地名
        level (str | None): 查询区域的行政等级，包括province、city、county三级。默认值为None，当为None时不区分查找范围，因此很可能出现重名错误

    Returns:
        str | None: 上级行政区全称
    """
    if name in ZXCITIES.keys():
        new_name = ZXCITIES[name]
        obj_area = Addr(new_name, level="province")
    else:
        obj_area = Addr(name, level=level)
    return obj_area._belongs_to()


def coordinate(name: str, level: str | None = None) -> tuple[float, float]:
    """根据区域名得到其行政中心的坐标

    Args:
        name (str): 行政区名称
        level (str | None): 查询区域的行政等级，包括province、city、county三级。默认值为None，当为None时不区分查找范围，因此很可能出现重名错误

    Returns:
        tuple[float, float]: 行政中心经纬度
    """
    if name in ZXCITIES.keys():
        new_name = ZXCITIES[name]
        obj_area = Addr(new_name, level="province")
    else:
        obj_area = Addr(name, level=level)
    lat, lon = obj_area._coordinate()
    return lat, lon


def haversine(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    """利用两个地点的经纬度计算球面距离

    Args:
        lat1 (float): 地区1的纬度
        lon1 (float): 地区1的经度
        lat2 (float): 地区2的纬度
        lon2 (float): 地区2的经度

    Returns:
        float: 球面距离，单位：km
    """
    # 将十进制度数转化为弧度
    lat1, lon1, lat2, lon2 = map(radians, [lat1, lon1, lat2, lon2])  # radians：将角度转化为弧度

    # haversine（半正矢）公式
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    r = 6371  # 地球平均半径，单位为公里
    return c * r


def dist(
    name1: str, name2: str, level1: str | None = None, level2: str | None = None
) -> float:
    """利用两个地区的城市名称计算球面距离

    Args:
        name1 (str): 地区1的城市名称
        name2 (str): 地区2的城市名称
        level1 (str | None): 查询地区1的行政等级，包括province、city、county三级。
        level2 (str | None): 查询地区2的行政等级，包括province、city、county三级。

    Returns:
        float: 球面距离，单位：km

    Note:
        level1、level2默认值为None，当为None时不区分查找范围，因此很可能出现重名错误。
    """
    city1_lat, city1_lon = coordinate(name1, level=level1)
    city2_lat, city2_lon = coordinate(name2, level=level2)
    return haversine(city1_lat, city1_lon, city2_lat, city2_lon)



import itertools
import numpy as np
from .non_dimension import toone


def __variability(data_origin: np.ndarray) -> np.ndarray:  # critic方法指标变异性数据
    data = toone(data_origin.copy(), mode='0')
    assert isinstance(data, np.ndarray)
    m, n = data.shape

    variabilities = []
    for j in range(n):
        ave_x = data[:, j].mean()
        diff = np.array(data[:, j] - ave_x)
        sum_var = np.sum(diff**2 / (m - 1))
        variabilities.append(sum_var**0.5)
    return np.array(variabilities)


def __conflict(data_origin: np.ndarray) -> np.ndarray:  # critic方法指标冲突性数据
    corr_matrix = np.corrcoef(data_origin, rowvar=False)
    conflicts = []
    p, q = corr_matrix.shape
    conflicts.extend(sum((1 - corr_matrix[i, j]) for i in range(p)) for j in range(q))

    return np.array(conflicts)


def critic(data_origin: np.ndarray) -> np.ndarray:
    """通过所提供数据计算critic权重

    Args:
        data_origin (np.ndarray): 待计算权重的数据

    Returns:
        np.ndarray: critic权重数组
    """
    info1 = __variability(data_origin)
    info2 = __conflict(data_origin)
    information = np.array(info1) * np.array(info2)
    information[np.isnan(information)] = 0

    _, q = data_origin.shape
    sum_info = information.sum()
    if sum_info == 0:
        return np.ones(q) / q
    weights = [information[c] / sum_info for c in range(q)]
    return np.array(weights)


def ewm(data_origin: np.ndarray) -> np.ndarray:
    """通过所提供数据计算entropy weight method(ewm)权重

    Args:
        data_origin (np.ndarray): 待计算权重的数据

    Returns:
        np.ndarray: ewm权重数组
    """
    data = toone(data_origin.copy(), mode='0')
    assert isinstance(data, np.ndarray)
    m, n = data.shape
    entropy = []

    for j in range(n):
        data[:, j] = data[:, j] / np.sum(data[:, j])
        np.place(data[:, j], data[:, j] == 0, 1)
        ej = -np.log(1 / m) * np.sum(data[:, j] * np.log(data[:, j]))
        entropy.append(ej)

    weights = [(1 - entropy[c]) / (m - np.sum(entropy)) for c in range(n)]
    return np.array(weights)


def stddev(data_origin: np.ndarray) -> np.ndarray:
    """通过所提供数据计算standard deviation(stddev)权重

    Args:
        data_origin (np.ndarray): 待计算权重的数据

    Returns:
        np.ndarray: stddev权重数组
    """
    data = toone(data_origin.copy(), mode='0')
    assert isinstance(data, np.ndarray)
    _, n = data.shape

    info = [np.std(data[:, j]) for j in range(n)]
    if np.sum(info) == 0:
        return np.ones(n) / n
    weights = [(i / np.sum(info)) for i in info]
    return np.array(weights)


def gini(data_origin: np.ndarray) -> np.ndarray:
    """通过所提供数据计算基尼系数法权重

    Args:
        data_origin (np.ndarray): 待计算权重的数据

    Returns:
        np.ndarray: 基尼系数法权重数组
    """
    data = data_origin.copy()
    m, n = data.shape
    Gini = []
    for j in range(n):
        gini_j = [
            np.abs(data[u, j] - data[v, j])
            for u, v in itertools.product(range(m), range(m))
        ]
        Gini.append(np.sum(gini_j) / (m**2 - m))
    return np.array(Gini / np.sum(Gini))



import numpy as np
from .non_dimension import toone


def topsis(
    data_origin: np.ndarray, weights: list[int | float] | np.ndarray
) -> np.matrix:
    """计算优劣解距离法得分矩阵，weights为权重矩阵。

    Args:
        data_origin (np.ndarray): 待计算数据
        weights (np.ndarray): 权重数组

    Returns:
        np.matrix: 若参数无误，返回得分数据，否则返回None
    """
    data = data_origin.copy()
    data = toone(data, mode='3')
    assert isinstance(data, np.ndarray)
    m, n = data.shape

    empty_matrix1 = np.empty((m, n))
    empty_matrix2 = np.empty((m, n))
    z_max = [data[:, j].max() for j in range(n)]
    z_min = [data[:, j].min() for j in range(n)]

    for j in range(n):
        empty_matrix1[:, j] = weights[j] * (z_max[j] - data[:, j]) ** 2
        empty_matrix2[:, j] = weights[j] * (z_min[j] - data[:, j]) ** 2

    d1: np.ndarray = np.sqrt(empty_matrix1.sum(axis=1))
    d2: np.ndarray = np.sqrt(empty_matrix2.sum(axis=1))
    result: np.ndarray = d2 / (d1 + d2)
    return np.mat(result.reshape(result.shape[0], 1))


def rsr(data_origin: np.ndarray, weights: list[int | float] | np.ndarray) -> np.matrix:
    """计算整次秩和比得分矩阵，weights为权重矩阵。

    Args:
        data_origin (np.ndarray): 待计算数据
        weights (np.ndarray): 权重数组

    Returns:
        np.matrix: 若参数无误，返回得分数据，否则返回None
    """
    data = data_origin.copy()
    m, n = data.shape

    rsr_matrix = np.empty((m, n))
    for q in range(n):
        compare_list: np.ndarray = np.sort(data[:, q])
        rsr_matrix[:, q] = np.searchsorted(compare_list, data[:, q])
    return rsr_matrix * np.mat(weights).T / m  # type: ignore


def ni_rsr(
    data_origin: np.ndarray, weights: list[int | float] | np.ndarray
) -> np.matrix:
    """计算非整次秩和比得分矩阵，weights为权重矩阵。

    Args:
        data_origin (np.ndarray): 待计算数据
        weights (np.ndarray): 权重数组

    Returns:
        np.matrix: 若参数无误，返回得分数据，否则返回None
    """
    data = data_origin.copy()
    m, n = data.shape

    rsr_matrix = np.empty((m, n))
    for q in range(n):
        max_v = data[:, q].max()
        min_v = data[:, q].min()
        rsr_matrix[:, q] = 1 + ((m - 1) * (data[:, q] - min_v) / (max_v - min_v))
    return rsr_matrix * np.mat(weights).T / m  # type: ignore



from jieba import lcut
import jieba
import polars as pl
from .special import SIMILAR_NAMES, SAMENAMES
from .packagepath import full_path

try:
    data = pl.read_csv(full_path())
except Exception:
    data = pl.read_csv("src/cp_lookup/attachment/adcodes.csv")

data["adcode"] = data["adcode"].apply(str)

SPECIAL_WORDS = [area for area in data["name"] if len(lcut(area)) != 1]
for word in SPECIAL_WORDS:
    jieba.add_word(word)


class AreaNameError(Exception):
    """自定义异常类"""

    def __init__(self, error_info):
        super().__init__(self)
        self.error_info = error_info

    def __str__(self):
        return self.error_info


def _prov_codes() -> list[str]:
    adcodes = data["adcode"]
    return [code for code in adcodes if str.endswith(code, "0" * 10)]


def _city_codes() -> list[str]:
    adcodes = data["adcode"]
    prov_codes = _prov_codes()
    return [
        code
        for code in adcodes
        if (str.endswith(code, "0" * 8)) & (code not in prov_codes)
    ]


def _county_codes() -> list[str]:
    adcodes = data["adcode"]
    prov_codes = _prov_codes()
    city_codes = _city_codes()
    return [
        code for code in adcodes if (code not in prov_codes) & (code not in city_codes)
    ]


PROVINCECODE: list[str] = _prov_codes()
CITYCODE: list[str] = _city_codes()
COUNTYCODE: list[str] = _county_codes()
NOTPROVINCE: list[str] = CITYCODE + COUNTYCODE
NOTCOUNTY: list[str] = PROVINCECODE + CITYCODE


def _level_choose(level: str | None) -> pl.DataFrame:
    options = ["province", "city", "county", "not province", "not county", None]
    assert level in options, "错误的等级选项！！！"
    match level:
        case "province":
            return data[data["adcode"].is_in(PROVINCECODE)]
        case "city":
            return data[data["adcode"].is_in(CITYCODE)]
        case "county":
            return data[data["adcode"].is_in(COUNTYCODE)]
        case "not province":
            return data[data["adcode"].is_in(NOTPROVINCE)]
        case "not county":
            return data[data["adcode"].is_in(NOTCOUNTY)]
        case _:
            return data


def _num_one(info: list[str], level_data: pl.DataFrame):
    addr = info[0]
    if len(level_data[level_data["name"].str.contains(addr)]) != 1:
        raise AreaNameError("地名重复或有误！！！")
    else:
        return level_data[level_data["name"].str.contains(addr)]


def _num_two(info: list[str], level_data: pl.DataFrame):
    father, addr = info
    if father == addr:
        addr_name = SAMENAMES[addr][1]
        return data[data["name"] == addr_name]
    else:
        fdata = _level_choose("not county")
        fcode = fdata[fdata["name"].str.contains(father)][0, "adcode"]
        addr_codes = level_data[level_data["name"].str.contains(addr)]["adcode"]
        for code in addr_codes:
            if (code[:4] == fcode[:4]) | (code[:2] == fcode[:2]):
                return level_data[level_data["adcode"] == code]


def _num_three(info: list[str], level_data: pl.DataFrame):
    gdfather, father, addr = info
    gddata = _level_choose("province")
    fdata = _level_choose("city")
    ldata = _level_choose("county")
    if len(level_data[level_data["name"].str.contains(addr)]) == 1:
        return level_data[level_data["name"].str.contains(addr)]
    gdcode = gddata[gddata["name"].str.contains(gdfather)][0, "adcode"]
    fdcode = fdata[fdata["name"].str.contains(father)][0, "adcode"]
    addr_codes = ldata[ldata["name"].str.contains(addr)]["adcode"]
    for code in addr_codes:
        if code[:2] == fdcode[:2] == gdcode[:2]:
            return level_data[level_data["adcode"] == code]


class Addr:
    """以区域名构建一个地区的类，包含该地区的adcode、区域名、经纬度等信息。"""

    def __init__(self, name: str, adcode: int | None = None, level: str | None = None):
        if adcode:
            self.addr = data[data["adcode"] == adcode]
        elif name in SIMILAR_NAMES.keys():
            self.addr = data[data["adcode"] == SIMILAR_NAMES.get(name)]
        else:
            obj_area_data = _level_choose(level) if level else data
            info = lcut(name)
            nums = len(info)
            match nums:
                case 1:
                    self.addr = _num_one(info, obj_area_data)
                case 2:
                    self.addr = _num_two(info, obj_area_data)
                case 3:
                    self.addr = _num_three(info, obj_area_data)

    def _belongs_to(self) -> str | None:
        assert isinstance(self.addr, pl.DataFrame)
        code = self.addr[0, "adcode"]
        fcode = (
            code[:2] + "0" * 10 if str.endswith(code, "0" * 8) else (code[:4] + "0" * 8)  # type: ignore
        )
        fdata = data[data["adcode"] == fcode]
        return fdata[0, "name"]  # type: ignore

    def _coordinate(self) -> tuple[float, float]:
        lat = self.addr[0, "latitude"]  # type: ignore
        lon = self.addr[0, "longitude"]  # type: ignore
        return lat, lon  # type: ignore



from functools import reduce
import numpy as np


############################################################################################
# 辅助函数区域
############################################################################################


def fn(x: int | float, y: int | float) -> int | float:  # 辅助函数，用于向量归一化计算
    return x**2 + y**2


def fn2(
    x: int | float, low: int | float, high: int | float, M: int | float
) -> int | float:  # 辅助函数，用于计算适度性指标转化为极大型指标的公式
    assert low <= high
    if x < low:
        return 1 - (low - x) / M
    elif x > high:
        return 1 - (x - high) / M
    else:
        return 1


vfn2 = np.vectorize(fn2)
##############################################################################################


def tiny_convert(
    ndarray: np.ndarray, mode: str, change_list: list[int]
) -> np.ndarray | None:
    """极小型指标转化为极大型指标

    Args:
        ndarray (np.ndarray): 待转化数据
        mode (str): 转化模式。mode='0'时为1/x模式，mode='1'时为max(x)-x模式
        change_list (list[int]): 待转化列下标组成的列表

    Returns:
        np.ndarray | None: 若参数无误，返回转化后的数据组，否则返回None
    """
    df1 = ndarray.copy().astype(float)
    if mode == '0':
        for j in change_list:
            df1[:, j] = 1 / df1[:, j]
        return df1

    elif mode == '1':
        for j in change_list:
            df1[:, j] = df1[:, j].max() - df1[:, j]
        return df1

    else:
        print('请正确地选择模式')


def middle_convert(
    ndarray: np.ndarray, change_list: list[int], best_value: list[int | float]
) -> np.ndarray | None:
    """中间型指标转化为极大型指标

    Args:
        ndarray (np.ndarray): 待转化数据
        change_list (list[int]): 待转化列下标组成的列表
        best_value (list[int|float]): 待转化列的最优值组成的列表。由于每个列的最优值不同，所以需要提供

    Returns:
        np.ndarray | None: 若参数无误，返回转化后的数据组，否则返回None
    """
    if len(change_list) != len(best_value):
        print("请检查可变列表和最佳值列表的元素数量")
    else:
        copy_matrix = ndarray.copy().astype(float)
        for j in change_list:
            M = max(np.abs(copy_matrix[:, j] - best_value[change_list.index(j)]))
            copy_matrix[:, j] = (
                1 - np.abs(copy_matrix[:, j] - best_value[change_list.index(j)]) / M
            )
        return copy_matrix


def moderate_convert(
    ndarray: np.ndarray,
    change_list: list[int],
    low_limit: list[int | float],
    high_limit: list[int | float],
) -> np.ndarray | None:
    """适度性指标转化为极大型指标

    Args:
        ndarray (np.ndarray): 待转化数据
        change_list (list[int]): 待转化列下标组成的列表
        low_limit (list[int|float]): 待转化列的最优区间下限组成的列表。由于每个列的最优区间下限不同，所以需要提供
        high_limit (list[int|float]): 待转化列的最优区间上限组成的列表。由于每个列的最优区间上限不同，所以需要提供

    Returns:
        np.ndarray | None: 若参数无误，返回转化后的数据组，否则返回None
    """
    if len(low_limit) == len(high_limit) == len(change_list):
        df2 = ndarray.copy().astype(float)
        for j in change_list:
            a = low_limit[change_list.index(j)]
            b = high_limit[change_list.index(j)]
            assert a <= b
            M = max(a - min(df2[:, j]), max(df2[:, j]) - b)
            df2[:, j] = vfn2(df2[:, j], a, b, M)
        return df2

    else:
        print("请检查可变列表与上下界列表的元素数量")


def toone(origin_array: np.ndarray, mode: str) -> np.ndarray | None:
    """多种矩阵归一化方法

    Args:
        ndarray (np.ndarray): 待转化数据
        mode (str): 转化模式。mode='0'时为归一化，mode='1'时为平均归一化，mode='2'时为标准化，mode='3'时为向量归一化

    Returns:
        np.ndarray | None: 若参数无误，返回转化后的数据组，否则返回None
    """
    ndarray = origin_array.copy().astype(float)
    m, n = ndarray.shape
    copy_matrix = np.empty((m, n))
    match mode:
        case '0':
            for j in range(n):
                mmax = ndarray[:, j].max()
                mmin = ndarray[:, j].min()
                copy_matrix[:, j] = (
                    (ndarray[:, j] - mmin) / (mmax - mmin) if mmax != mmin else 1
                )
            return copy_matrix
        case '1':
            for j in range(n):
                mmean = ndarray[:, j].mean()
                mmax = ndarray[:, j].max()
                mmin = ndarray[:, j].min()
                copy_matrix[:, j] = (
                    (ndarray[:, j] - mmean) / (mmax - mmin) if mmax != mmin else 0
                )
            return copy_matrix
        case '2':
            for j in range(n):
                mmean = ndarray[:, j].mean()
                mstd = ndarray[:, j].std()
                copy_matrix[:, j] = (ndarray[:, j] - mmean) / mstd if mstd != 0 else 0
            return copy_matrix
        case '3':
            for j in range(n):
                vec_length = np.sqrt(np.array(reduce(fn, ndarray[:, j])))
                copy_matrix[:, j] = ndarray[:, j] / vec_length
            return copy_matrix



from cp_lookup import belongs_to, lookup, dist, coordinate


def test_belongs_to():
    assert belongs_to("盘锦") == "辽宁省"
    assert belongs_to("青海海南") == "青海省"


def test_lookup():
    assert lookup("北京朝阳") == "朝阳区"
    assert lookup("青海海南") == "海南藏族自治州"


def test_dist():
    assert dist("铁岭市", "自贡市") == 2221.108451374754


def test_coordinate():
    assert coordinate("齐齐哈尔") == (47.354348, 123.918186)
